@@ -1,4 +1,4 @@
-# $Id: CDDB.pm,v 1.24 2001/05/03 20:30:38 troc Exp troc $
+# $Id: CDDB.pm,v 1.25 2002/12/22 04:34:00 troc Exp $
 # Documentation and Copyright exist after __END__
 
 package CDDB;
@@ -8,7 +8,7 @@ use strict;
 use vars qw($VERSION);
 use Carp;
 
-$VERSION = '1.07';
+$VERSION = '1.11';
 
 BEGIN {
   if ($^O eq 'MSWin32') {
@@ -29,7 +29,11 @@ use Sys::Hostname;
 # others anyway.
 
 my @cddbp_hosts =
-  ( [ 'freedb.freedb.org' => 8880 ],
+  ( [ 'localhost'         => 8880 ],
+    [ 'freedb.freedb.org' => 8880 ],
+    [ 'us.freedb.org',    => 8880 ],
+    [ 'ca.freedb.org',    => 8880 ],
+    [ 'ca2.freedb.org',   => 8880 ],
     [ 'uk.freedb.org'     => 8880 ],
     [ 'no.freedb.org'     => 8880 ],
     [ 'de.freedb.org'     => 8880 ],
@@ -37,8 +41,8 @@ my @cddbp_hosts =
     [ 'freedb.freedb.de'  => 8880 ],
   );
 
-# Determine whether we can submit changes by e-mail.  TODO: HTTP
-# submissions.
+#------------------------------------------------------------------------------
+# Determine whether we can submit changes by e-mail.
 
 my $imported_mail = 0;
 eval {
@@ -48,9 +52,19 @@ eval {
   $imported_mail = 1;
 };
 
-# Send a command.  If we're not connected, try to connect
-# first. Returns 1 if the command is sent ok; 0 if there was a
-# problem.
+#------------------------------------------------------------------------------
+# Determine whether we can use HTTP for requests and submissions.
+
+my $imported_http = 0;
+eval {
+  require LWP;
+  require HTTP::Request;
+  $imported_http = 1;
+};
+
+#------------------------------------------------------------------------------
+# Send a command.  If we're not connected, try to connect first.
+# Returns 1 if the command is sent ok; 0 if there was a problem.
 
 sub command {
   my $self = shift;
@@ -69,6 +83,7 @@ sub command {
   return 1;
 }
 
+#------------------------------------------------------------------------------
 # Retrieve a line from the server.  Uses a buffer to allow for
 # ungetting lines.  Returns the next line or undef if there is a
 # problem.
@@ -116,6 +131,7 @@ sub getline {
   return $line;
 }
 
+#------------------------------------------------------------------------------
 # Receive a server response, and parse it into its numeric code and
 # text message.  Return the code's first character, which usually
 # indicates the response class (ok, error, information, warning,
@@ -137,6 +153,7 @@ sub response {
   substr($code, 0, 1);
 }
 
+#------------------------------------------------------------------------------
 # Accessors to retrieve the last response() call's code and text
 # separately.
 
@@ -150,6 +167,7 @@ sub text {
   $self->{response_text};
 }
 
+#------------------------------------------------------------------------------
 # Helper to print stuff for debugging.
 
 sub debug_print {
@@ -163,6 +181,7 @@ sub debug_print {
   print STDERR $text, "\n";
 }
 
+#------------------------------------------------------------------------------
 # Read data until it's terminated by a single dot on its own line.
 # Two dots at the start of a line are replaced by one.  Returns an
 # ARRAY reference containing the lines received, or undef on error.
@@ -181,6 +200,7 @@ sub read_until_dot {
   \@lines;
 }
 
+#------------------------------------------------------------------------------
 # Create an object to represent one or more cddbp sessions.
 
 sub new {
@@ -218,7 +238,7 @@ sub new {
 
   # Choose a particular cddbp port.
   my $port = $param{Port};
-  $port = 0 unless defined $port;
+  $port = 8880 unless $port;
 
   # Choose a particular cddbp submission address.
   my $submit_to = $param{Submit_Address};
@@ -232,6 +252,10 @@ sub new {
   my $client_version = $param{Client_Version};
   $client_version = $VERSION unless defined $client_version;
 
+  # Change the cddbp protocol level.
+  my $cddb_protocol = $param{Protocol_Version};
+  $cddb_protocol = 1 unless defined $cddb_protocol;
+
   # Mac Freaks Got Spaces!  Augh!
   $login =~ s/\s+/_/g;
 
@@ -246,6 +270,7 @@ sub new {
     debug         => $debug,
     host          => $host,
     port          => $port,
+    cddb_protocol => $cddb_protocol,
     lines         => [],
     frame         => '',
     response_code => '000',
@@ -255,6 +280,7 @@ sub new {
   $self;
 }
 
+#------------------------------------------------------------------------------
 # Disconnect from a cddbp server.  This is needed sometimes when a
 # server decides a session has performed enough requests.
 
@@ -270,6 +296,7 @@ sub disconnect {
   }
 }
 
+#------------------------------------------------------------------------------
 # Connect to a cddbp server.  Connecting and disconnecting are done
 # transparently and are performed on the basis of need.  Furthermore,
 # this routine will cycle through servers until one connects or it has
@@ -307,16 +334,19 @@ HOST:
       # TODO: give bad hosts extra chances in case there are transient
       # network problems.
       if ($self->{host} eq '') {
+
+        # None of the servers worked.  Time to leave.
+        unless (@cddbp_hosts) {
+          $self->debug_print( 0, "--- all cddbp servers failed to answer" );
+          warn "No cddb protocol servers answer.  Is your network OK?\n"
+            unless $self->{debug};
+          return undef;
+        }
+
         $cddbp_host = shift(@cddbp_hosts);
         ($self->{host}, $self->{port}) = @$cddbp_host;
       }
 
-      # None of the servers worked.  Time to leave.
-      unless (length $self->{host}) {
-        $self->debug_print( 0, "--- all cddbp servers failed to answer" );
-        return undef;
-      }
-
       # Assign the host we selected, and attempt a connection.
       $self->debug_print
         ( 0,
@@ -386,6 +416,19 @@ HOST:
       next HANDSHAKE;
     }
 
+    # Set the protocol level.
+    if ($self->{cddb_protocol} != 1) {
+      $self->command( 'proto', $self->{cddb_protocol} );
+      $code = $self->response();
+      if ($code != 2) {
+        $self->debug_print( 0, "--- can't set protocol level ",
+                            $self->{cddb_protocol}, ' ',
+                            $self->code(), ' ', $self->text()
+                          );
+        return undef;
+      }
+    }
+
     # If we get here, everything succeeded.
     return 1;
   }
@@ -398,9 +441,10 @@ sub DESTROY {
   $self->disconnect();
 }
 
-#------------------------------------------------------------------------------
+###############################################################################
 # High-level cddbp functions.
 
+#------------------------------------------------------------------------------
 # Get a list of available genres.  Returns an array of genre names, or
 # undef on failure.
 
@@ -424,6 +468,7 @@ sub get_genres {
   return undef;
 }
 
+#------------------------------------------------------------------------------
 # Calculate a cddbp ID based on a text table of contents.  The text
 # format was chosen because it was straightforward and easy to
 # generate.  In a scalar context, this returns just the cddbp ID.  In
@@ -492,6 +537,26 @@ sub calculate_id {
   return $id;
 }
 
+#------------------------------------------------------------------------------
+# Parse cdinfo's output so calculate_id() can eat it.
+
+sub parse_cdinfo {
+  my ($self, $command) = @_;
+  open(FH, $command) or croak "could not open `$command': $!";
+
+  my @toc;
+  while (<FH>) {
+    if (/(\d+):\s+(\d+):(\d+):(\d+)/) {
+      my @track = ($1,$2,$3,$4);
+      $track[0] = 999 if /leadout/;
+      push @toc, "@track";
+    }
+  }
+  close FH;
+  return @toc;
+}
+
+#------------------------------------------------------------------------------
 # Get a list of discs that match a particular CD's table of contents.
 # This accepts the TOC information as returned by calculate_id().  It
 # will also accept information in mp3 format, but I forget what that
@@ -525,7 +590,7 @@ ATTEMPT:
     # Send a cddbp query command.
     $self->command( 'cddb query', $id, $track_count,
                     $offsets_string, $total_seconds
-                  );
+                  ) or return undef;
 
     # Get the response.  Try again if the server is temporarly
     # unavailable.
@@ -550,7 +615,9 @@ ATTEMPT:
   return () if $self->code() == 202;
 
   # Multiple matching discs.
-  if ($self->code() == 211) {
+  # 210 Found exact matches, list follows (...)   [proto>=4]
+  # 211 Found inexact matches, list follows (...) [proto>=1]
+  if ($self->code() == 210 or $self->code() == 211) {
     my $discs = $self->read_until_dot();
     return undef unless defined $discs;
 
@@ -570,6 +637,7 @@ ATTEMPT:
   return undef;
 }
 
+#------------------------------------------------------------------------------
 # A little helper to combine list-context calculate_id() with
 # get_discs().
 
@@ -582,6 +650,19 @@ sub get_discs_by_toc {
   @discs;
 }
 
+#------------------------------------------------------------------------------
+# A little helper to get discs from an existing query string.
+# Contributed by Ron Grabowski.
+
+sub get_discs_by_query {
+  my ($self, $query) = @_;
+  my (undef, undef, $cddbp_id, $tracks, @offsets) = split /\s+/, $query;
+  my $total_seconds = pop @offsets;
+  my @discs = $self->get_discs($cddbp_id, \@offsets, $total_seconds);
+  return @discs;
+}
+
+#------------------------------------------------------------------------------
 # Retrieve the database record for a particular genre/id combination.
 # Returns a moderately complex hashref representing the cddbp record,
 # or undef on failure.
@@ -616,9 +697,14 @@ sub get_disc_details {
   # Parse that puppy.
 
   my @track_file = @$track_file;
-  my %details = ( offsets => [ ] );
+  my %details = ( offsets => [ ],
+                  seconds => [ ],
+                );
   my $state = 'beginning';
   foreach my $line (@track_file) {
+    # Keep returned so-called xmcd record...
+    $details{xmcd_record} .= $line . "\n";
+
     if ($state eq 'beginning') {
       if ($line =~ /track\s*frame\s*off/i) {
         $state = 'offsets';
@@ -632,9 +718,10 @@ sub get_disc_details {
         next;
       }
       $state = 'headers';
-      next;
+      # This passes through on purpose.
     }
 
+    # This is not an elsif on purpose.
     if ($state eq 'headers') {
       if ($line =~ /^\#/) {
         $line =~ s/\s+/ /g;
@@ -671,12 +758,37 @@ sub get_disc_details {
     }
   }
 
+  # Translate disc offsets into seconds.  This builds a virtual track
+  # 0, which is the time from the beginning of the disc to the
+  # beginning of the first song.  That time's used later to calculate
+  # the final track's length.
+
+  my $last_offset = 0;
+  foreach (@{$details{offsets}}) {
+    push( @{$details{seconds}},
+          int(($_ - $last_offset) / 75)
+        );
+    $last_offset = $_;
+  }
+
+  # Create the final track length from the disc length.  Remove the
+  # virtual track 0 in the process.
+
+  my $disc_length = $details{"disc length"};
+  $disc_length =~ s/ .*$//;
+
+  my $first_start = shift @{$details{seconds}};
+  push( @{$details{seconds}},
+        $disc_length - int($details{offsets}->[-1] / 75) + 1 - $first_start
+      );
+
   \%details;
 }
 
-#------------------------------------------------------------------------------
+###############################################################################
 # Evil voodoo e-mail submission stuff.
 
+#------------------------------------------------------------------------------
 # Return true/false whether the libraries needed to submit discs are
 # present.
 
@@ -685,6 +797,7 @@ sub can_submit_disc {
   $imported_mail;
 }
 
+#------------------------------------------------------------------------------
 # Build an e-mail address, and return it.  Caches the last built
 # address, and returns that on subsequent calls.
 
@@ -694,6 +807,7 @@ sub get_mail_address {
   return $self->{mail_from} = $self->{login} . '@' . $self->{hostname};
 }
 
+#------------------------------------------------------------------------------
 # Build an e-mail host, and return it.  Caches the last built e-mail
 # host, and returns that on subsequent calls.
 
@@ -730,7 +844,7 @@ sub submit_disc {
     $self->{hostname} = &hostname() or croak "can't get hostname: $!";
   }
 
-  # Validate the required submission fields.
+  # Validate the required submission fields.  XXX Duplicated code.
   (exists $params{Genre})       or croak "submit_disc needs a Genre";
   (exists $params{Id})          or croak "submit_disc needs an Id";
   (exists $params{Artist})      or croak "submit_disc needs an Artist";
@@ -847,16 +961,16 @@ CDDB.pm - a high-level interface to cddb protocol servers (freedb and CDDB)
 
   use CDDB;
 
-  ### connect to the cddbp server
-  my $cddbp = new CDDB( Host  => 'freedb.freedb.org',    # default
-                        Port  => 8880,                   # default
-                        Login => $login_id,              # defaults to %ENV's
+  ### Connect to the cddbp server.
+  my $cddbp = new CDDB( Host  => 'freedb.freedb.org', # default
+                        Port  => 8880,                # default
+                        Login => $login_id,           # defaults to %ENV's
                       ) or die $!;
 
-  ### retrieve known genres
+  ### Retrieve known genres.
   my @genres = $cddbp->get_genres();
 
-  ### calculate cddbp ID based on MSF info
+  ### Calculate cddbp ID based on MSF info.
   my @toc = ( '1    0  2 37',           # track, CD-i MSF (space-delimited)
               '999  1 38 17',           # lead-out track MSF
               '1000 0  0 Error!',       # error track (don't include if ok)
@@ -868,22 +982,23 @@ CDDB.pm - a high-level interface to cddb protocol servers (freedb and CDDB)
       $total_seconds  # total play time, in seconds (for cddbp queries)
      ) = $cddbp->calculate_id(@toc);
 
-  ### query discs based on cddbp ID and other information
+  ### Query discs based on cddbp ID and other information.
   my @discs = $cddbp->get_discs($cddbp_id, $track_offsets, $total_seconds);
   foreach my $disc (@discs) {
     my ($genre, $cddbp_id, $title) = @$disc;
   }
 
-  ### query disc details (usually done with get_discs() information)
+  ### Query disc details (usually done with get_discs() information).
   my $disc_info     = $cddbp->get_disc_details($genre, $cddbp_id);
   my $disc_time     = $disc_info->{'disc length'};
   my $disc_id       = $disc_info->{discid};
   my $disc_title    = $disc_info->{dtitle};
   my @track_offsets = @{$disc_info->{offsets}};
+  my @track_seconds = @{$disc_info->{seconds}};
   my @track_titles  = @{$disc_info->{ttitles}};
   # other information may be returned... explore!
 
-  ### submit a disc (via e-mail, requires MailTools)
+  ### Submit a disc via e-mail. (Requires MailTools)
 
   die "can't submit a disc (no mail modules; see README)"
     unless $cddbp->can_submit_disc();
@@ -941,6 +1056,10 @@ B<Host> and B<Port> describe the cddbp server to connect to.  These
 default to 'freedb.freedb.org' and 8880, which is a multiplexor for
 all the other freedb servers.
 
+B<Protocol_Version> sets the cddbp version to use.  CDDB.pm will not
+connect to servers that don't support the version specified here.  The
+requested protocol version defaults to 1 if omitted.
+
 B<Login> is the login ID you want to advertise to the cddbp server.
 It defaults to the login ID your computer assigns you, if that can be
 determined.  The default login ID is determined by the presence of a
@@ -982,6 +1101,9 @@ FreeBSD.  CDDB.pm's author has used external programs like dagrab to
 fetch the offsets.  Actual CDs aren't always necessary: the author has
 heard of people generating TOC information from mp3 file lengths.
 
+That said, see parse_cdinfo() for a routine to parse "cdinfo" output
+into a table of contents list suitable for calculate_id().
+
 calculate_id() accepts TOC information as a list of strings.  Each
 string contains four fields, separated by whitespace:
 
@@ -1072,9 +1194,13 @@ seconds.  The get_discs() function needs it.
 =item get_discs CDDBP_ID, TRACK_OFFSETS, TOTAL_SECONDS
 
 get_discs() asks the cddbp server for a summary of all the CDs
-matching a given cddbp ID, track offsets, and total playing time.  It
-accepts these in the same format returned by calculate_id(), so it's
-easy to pass the values from one function to the other.
+matching a given cddbp ID, track offsets, and total playing time.
+These values can be retrieved from calculade_id().
+
+  my @id_info       = $cddbp->calculate_id(@toc);
+  my $cddbp_id      = $id_info->[0];
+  my $track_offsets = $id_info->[3];
+  my $total_seconds = $id_info->[4];
 
 get_discs() returns an array of matching discs, each of which is
 represented by an array reference.  It returns an empty array if the
@@ -1108,6 +1234,16 @@ This function acts as a macro, combining calculate_id() and
 get_discs() calls into one function.  It takes the same parameters as
 calculate_id(), and it returns the same information as get_discs().
 
+=item get_discs_by_query QUERY_STRING
+
+Fetch discs by a pre-built cddbp query string.  Some disc querying
+programs report this string, and get_discs_by_query() is a convenient
+way to use that.
+
+Cddb protocol query strings look like:
+
+  cddb query $cddbp_id $track_count @offsets $total_seconds
+
 =item get_disc_details DISC_GENRE, CDDBP_ID
 
 This function fetches a disc's detailed information from a cddbp
@@ -1138,9 +1274,13 @@ returned by get_discs().
 
 $disc_details->{offsets}
 
-This is a reference to a list of absolute disc trac offsets, similar
+This is a reference to a list of absolute disc track offsets, similar
 to the TRACK_OFFSETS returned by calculate_id().
 
+$disc_details->{seconds}
+
+This is a reference to a list of track length, in seconds.
+
 $disc_details->{ttitles}
 
 This is a reference to a list of track titles.  These are the droids
@@ -1165,6 +1305,11 @@ This is the name and version of the software that submitted this cddbp
 record.  The main intention is to identify records that are submitted
 by broken software so they can be purged or corrected.
 
+$disc_details->{xmcd_record}
+
+The xmcd_record field contains a copy of the entire unprocessed cddbp
+response that generated all the other fields.
+
 =item can_submit_disc
 
 Returns true or false, depending on whether CDDB.pm has enough
@@ -1207,6 +1352,17 @@ can see it and have an opportunity to correct it.
                        Host => $mail_host,
                      );
 
+=item parse_cdinfo CDINFO_FILE
+
+Generates a table of contents suitable for calculate_id() based on the
+output of a program called "cdinfo".  CDINFO_FILE may either be a text
+file, or it may be the cdinfo program itself.
+
+  my @toc = parse_cdinfo("cdinfo.txt"); # read cdinfo.txt
+  my @toc = parse_cdinfo("cdinfo|");    # run cdinfo directly
+
+The table of contents can be passed directly to calculate_id().
+
 =item submit_disc DISC_DETAILS
 
 submit_disc() submits a disc record to a cddbp server.  Currently it
@@ -1276,6 +1432,8 @@ required, and CDDB.pm will try to figure one out on its own.  It will
 look at the SMTPHOSTS environment variable is not defined, it will try
 'mail' and 'localhost' before finally failing.
 
+=back
+
 =head1 PRIVATE METHODS
 
 Documented as being not documented.
@@ -1292,8 +1450,10 @@ done.
 
 =head1 CONTACT AND COPYRIGHT
 
-Copyright 1998-2001 Rocco Caputo E<lt>troc@netrus.netE<gt>.  All
-rights reserved.  This program is free software; you can redistribute
-it and/or modify it under the same terms as Perl itself.
+Copyright 1998-2002 Rocco Caputo.  All rights reserved.  This program
+is free software; you can redistribute it and/or modify it under the
+same terms as Perl itself.
+
+Rocco may be contacted at rcaputo@cpan.org.
 
 =cut
